@@ -18,44 +18,64 @@ along with this program if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
+import datetime
+
 from PyQt4.QtCore import QTimer
 from twisted.spread import pb
 from twisted.internet import reactor
 from twisted.internet.task import deferLater
-from twisted.internet.defer import Deferred, succeed, DeferredList
+from twisted.internet.defer import Deferred, succeed
+from twisted.internet.error import ReactorNotRunning
 from twisted.python.failure import Failure
-from util import logDebug, logException, logWarning, Duration
+from util import logDebug, logException, logWarning, Duration, m18nc, checkMemory
 from message import Message
-from common import InternalParameters, Debug
+from common import Internal, Debug
 from rule import Ruleset
 from meld import meldsContent
-from game import RemoteGame
+from game import PlayingGame
 from query import Transaction, Query
 from move import Move
 from animation import animate
 from intelligence import AIDefault
 from statesaver import StateSaver
+from player import PlayingPlayer
 
-class ClientTable(object):
-    """the table as seen by the client"""
-    # pylint: disable=R0902
-    # pylint: disable=R0913
-    # pylint says too many args, too many instance variables
-
-    rulesets = {} # all rulesets for all tables, key is hash
-
-    def __init__(self, client, tableid, gameid, status, ruleset, playOpen, autoPlay, wantedGame, playerNames,
-                 playersOnline, endValues):
-        self.client = client
+class Table(object):
+    """defines things common to both ClientTable and ServerTable"""
+    def __init__(self, tableid, ruleset, suspendedAt, running, playOpen, autoPlay, wantedGame):
         self.tableid = tableid
-        self.gameid = gameid
-        self.status = status
-        self.running = status == 'Running'
-        self.suspended = status.startswith('Suspended')
-        self.ruleset = ruleset
+        if isinstance(ruleset, Ruleset):
+            self.ruleset = ruleset
+        else:
+            self.ruleset = Ruleset.cached(ruleset)
+        self.suspendedAt = suspendedAt
+        self.running = running
         self.playOpen = playOpen
         self.autoPlay = autoPlay
         self.wantedGame = wantedGame
+
+    def status(self):
+        """a status string"""
+        result = ''
+        if self.suspendedAt:
+            result = m18nc('table status', 'Suspended')
+            result += ' ' + datetime.datetime.strptime(self.suspendedAt,
+                '%Y-%m-%dT%H:%M:%S').strftime('%c').decode('utf-8')
+        if self.running:
+            result += ' ' + m18nc('table status', 'Running')
+        return result or m18nc('table status', 'New')
+
+class ClientTable(Table):
+    """the table as seen by the client"""
+    # pylint: disable=too-many-instance-attributes
+    # pylint: disable=too-many-arguments
+
+    def __init__(self, client, tableid, ruleset, gameid, suspendedAt, running,
+                 playOpen, autoPlay, wantedGame, playerNames,
+                 playersOnline, endValues):
+        Table.__init__(self, tableid, ruleset, suspendedAt, running, playOpen, autoPlay, wantedGame)
+        self.client = client
+        self.gameid = gameid
         self.playerNames = playerNames
         self.playersOnline = playersOnline
         self.endValues = endValues
@@ -75,98 +95,76 @@ class ClientTable(object):
         return False
 
     def __str__(self):
-        return 'Table %d %s gameid=%s rules %s players %s online %s' % (self.tableid or 0,
-            self.status, self.gameid, self.ruleset.name,
-            ', '.join(self.playerNames), ', '.join(str(x) for x in self.playersOnline))
+        onlineNames = list(x for x in self.playerNames if self.isOnline(x))
+        offlineString = ''
+        offlineNames = list(x for x in self.playerNames if x not in onlineNames
+            and not x.startswith('Robot'))
+        if offlineNames:
+            offlineString = ' offline:' + ','.join(offlineNames)
+        return '%d(%s %s%s)' % (self.tableid, self.ruleset.name, ','.join(onlineNames), offlineString)
+
+    def __repr__(self):
+        return 'ClientTable(%s)' % str(self)
 
     def gameExistsLocally(self):
         """does the game exist in the data base of the client?"""
         assert self.gameid
         return bool(Query('select 1 from game where id=?', list([self.gameid])).records)
 
-    def humanPlayerNames(self):
-        """returns a list excluding robot players"""
-        return list(x for x in self.playerNames if not x.startswith('Robot '))
-
-    @staticmethod
-    def fromList(client, table):
-        """convert the tuples delivered by twisted into a more useful class objects
-        if tables share rulesets, the server sends them only once.
-        The other tables only get the hash of the ruleset."""
-        table = list(table) # we can replace items in lists but in not tuples
-        if isinstance(table[3], basestring):
-            # server only sent the hash
-            table[3] = ClientTable.rulesets[table[3]]
-        else:
-            # server sent full ruleset definition
-            table[3] = Ruleset.fromList(table[3])
-            ClientTable.rulesets[table[3].hash] = table[3]
-        return ClientTable(client, *table)  # pylint: disable=W0142
-
-class Client(pb.Referenceable):
+class Client(object, pb.Referenceable):
     """interface to the server. This class only implements the logic,
     so we can also use it on the server for robot clients. Compare
     with HumanClient(Client)"""
 
-    clients = []
+    def __del__(self):
+        self.game = None
 
-    def __init__(self, username=None, intelligence=AIDefault):
-        """username is something like Robot 1 or None for the game server"""
-        self.username = username
+    def __init__(self, name=None, intelligence=AIDefault):
+        """name is something like Robot 1 or None for the game server"""
+        self.name = name
         self.game = None
         self.intelligence = intelligence(self)
-        self.perspective = None # always None for a robot client
+        self.__connection = None
         self.tables = []
         self.table = None
         self.tableList = None
         self.sayable = {} # recompute for each move, use as cache
-        self.clients.append(self)
 
-    def tableById(self, tableid):
+    def delete(self):
+        """for better garbage collection"""
+        self.table = None
+        self.intelligence = None
+
+    @property
+    def connection(self):
+        """update main window title if needed"""
+        return self.__connection
+
+    @connection.setter
+    def connection(self, value):
+        """update main window title if needed"""
+        if self.__connection != value:
+            self.__connection = value
+            if Internal.field:
+                Internal.field.updateGUI()
+
+    def _tableById(self, tableid):
         """returns table with tableid"""
         for table in self.tables:
             if table.tableid == tableid:
                 return table
 
-    def tableByGameId(self, gameid):
-        """returns table with gameid"""
-        for table in self.tables:
-            if table.gameid == gameid:
-                return table
-
-    @staticmethod
-    def shutdownClients(exception=None):
-        """close connections to servers except maybe one"""
-        clients = Client.clients
-        def done():
-            """return True if clients is cleaned"""
-            return len(clients) == 0 or (exception and clients == [exception])
-        def disconnectedClient(dummyResult, client):
-            """now the client is really disconnected from the server"""
-            if client in clients:
-                # HumanClient.serverDisconnects also removes it!
-                clients.remove(client)
-        if isinstance(exception, Failure):
-            logException(exception)
-        for client in clients[:]:
-            if client.tableList:
-                client.tableList.hide()
-        if done():
-            return succeed(None)
-        deferreds = []
-        for client in clients[:]:
-            if client != exception and client.perspective:
-                deferreds.append(client.logout().addCallback(disconnectedClient, client))
-        return DeferredList(deferreds)
-
     @staticmethod
     def quitProgram(result=None):
         """now all connections to servers are cleanly closed"""
         if isinstance(result, Failure):
             logException(result)
-        InternalParameters.reactor.stop()
+        try:
+            Internal.reactor.stop()
+        except ReactorNotRunning:
+            pass
         StateSaver.saveAll()
-        field = InternalParameters.field
+        field = Internal.field
         if field:
             # if we have the ruleset editor visible, we get:
             # File "/hdd/pub/src/gitgames/kajongg/src/rulesetselector.py", line 194, in headerData
@@ -175,33 +173,34 @@ class Client(pb.Referenceable):
             # how can Qt get None? Same happens with QEvent, see statesaver.py
             if field.confDialog:
                 field.confDialog.hide()
+            field.hide() # do not make the user see the delay for stopping the reactor
         # we may be in a Deferred callback which would
         # catch sys.exit as an exception
         # and the qt4reactor does not quit the app when being stopped
+        Internal.quitWaitTime = 0
         QTimer.singleShot(10, Client.appquit)
 
     @staticmethod
     def appquit():
         """retry until the reactor really stopped"""
-        if InternalParameters.reactor.running:
+        if Internal.reactor.running:
+            Internal.quitWaitTime += 10
+            if Internal.quitWaitTime % 1000 == 0:
+                logDebug('waiting since %d seconds for reactor to stop' % (Internal.quitWaitTime // 1000))
             QTimer.singleShot(10, Client.appquit)
         else:
-            InternalParameters.app.quit()
+            if Internal.quitWaitTime > 1000:
+                logDebug('reactor stopped after %d seconds' % (Internal.quitWaitTime // 1000))
+            Internal.app.quit()
+            checkMemory()
 
-    def logout(self, dummyResult=None): # pylint: disable=R0201
+    def logout(self, dummyResult=None): # pylint: disable=no-self-use
         """virtual"""
         return succeed(None)
 
-    @apply
-    def host():
-        """the name of the host we are connected with"""
-        def fget(dummySelf):
-            return None # Client on the server
-        return property(**locals())
-
     def isRobotClient(self):
         """avoid using isinstance because that imports too much for the server"""
-        return bool(self.username)
+        return bool(self.name)
 
     @staticmethod
     def isHumanClient():
@@ -210,31 +209,37 @@ class Client(pb.Referenceable):
 
     def isServerClient(self):
         """avoid using isinstance because that imports too much for the server"""
-        return bool(not self.username)
-
-    def remote_tablesChanged(self, tables):
-        """update table list"""
-        self.tables = []
-        self.remote_newTables(tables)
+        return bool(not self.name)
 
     def remote_newTables(self, tables):
         """update table list"""
-        self.tables.extend(list(ClientTable.fromList(self, x) for x in tables))
+        newTables = list(ClientTable(self, *x) for x in tables) # pylint: disable=star-args
+        self.tables.extend(newTables)
+        if Debug.table:
+            logDebug('%s got new tables:%s' % (self.name, newTables))
 
-    def remote_replaceTable(self, table):
+    @staticmethod
+    def remote_serverRulesets(hashes):
+        """the server will normally send us hashes of rulesets. If
+        a hash is not known by us, tell the server so it will send the
+        full ruleset definition instead of the hash. It would be even better if
+        the server always only sends the hash and the client then says "I do
+        not know this ruleset, please send definition", but that would mean
+        more changes to the client code"""
+        return list(x for x in hashes if not Ruleset.hashIsKnown(x))
+
+    def tableChanged(self, table):
         """update table list"""
-        newClientTable = ClientTable.fromList(self, table)
-        oldTable = self.tableById(newClientTable.tableid)
-        if not oldTable:
-            # joining a suspended game changes tableid
-            oldTable = self.tableByGameId(newClientTable.gameid)
+        newTable = ClientTable(self, *table) # pylint: disable=star-args
+        oldTable = self._tableById(newTable.tableid)
         if oldTable:
             self.tables.remove(oldTable)
-            self.tables.append(newClientTable)
+            self.tables.append(newTable)
+        return oldTable, newTable
 
-    def remote_tableClosed(self, tableid, dummyMsg):
+    def remote_tableRemoved(self, tableid, dummyMsg, *dummyMsgArgs):
         """update table list"""
-        table = self.tableById(tableid)
+        table = self._tableById(tableid)
         if table:
             self.tables.remove(table)
 
@@ -243,11 +248,12 @@ class Client(pb.Referenceable):
         in our local data base - we want to use the same gameid everywhere"""
         with Transaction():
             query = Query('insert into game(id,seed) values(?,?)',
-                      list([gameid, self.host]), mayFail=True)
+                      list([gameid, self.connection.url]), mayFail=True)
             if query.rowcount() != 1:
                 return Message.NO
+        return Message.OK
 
-    def readyForGameStart(self, tableid, gameid, wantedGame, playerNames, shouldSave=True):
+    def readyForGameStart(self, tableid, gameid, wantedGame, playerNames, shouldSave=True, gameClass=None):
         """the game server asks us if we are ready. A robot is always ready."""
         def disagree(about):
             """do not bother to translate this, it should normally not happen"""
@@ -255,14 +261,16 @@ class Client(pb.Referenceable):
             msg = 'The data bases for game %s have different %s' % (self.game.seed, about)
             logWarning(msg)
             raise pb.Error(msg)
-        if self.isHumanClient():
-            assert not self.table
-            assert self.tables
-            self.table = self.tableById(tableid)
-            if not self.table:
-                raise pb.Error('client.readyForGameStart: tableid %d unknown' % tableid)
-        if self.table.suspended:
-            self.game = RemoteGame.loadFromDB(gameid, client=self)
+        if not self.table:
+            assert not self.isRobotClient()
+            self.table = self._tableById(tableid)
+        else:
+            assert self.isRobotClient()
+            # robot client instance: self.table is already set
+        if gameClass is None:
+            gameClass = PlayingGame
+        if self.table.suspendedAt:
+            self.game = gameClass.loadFromDB(gameid, self)
             self.game.assignPlayers(playerNames)
             if self.isHumanClient():
                 if self.game.handctr != self.table.endValues[0]:
@@ -273,7 +281,7 @@ class Client(pb.Referenceable):
                         disagree('balances for wind %s: Server:%s, Client:%s' % (
                             player.wind, self.table.endValues[1][player.wind], player.balance))
         else:
-            self.game = RemoteGame(playerNames, self.table.ruleset,
+            self.game = gameClass(playerNames, self.table.ruleset,
                 shouldSave=shouldSave, gameid=gameid, wantedGame=wantedGame, client=self,
                 playOpen=self.table.playOpen, autoPlay=self.table.autoPlay)
         self.game.prepareHand()
@@ -286,36 +294,40 @@ class Client(pb.Referenceable):
             self.game.rotateWinds()
         self.game.prepareHand()
 
+    def __delayAnswer(self, result, delay, delayStep):
+        """try again, may we chow now?"""
+        noClaimCount = 0
+        delay += delayStep
+        for move in self.game.lastMoves():
+            # latest move first
+            if move.message == Message.Discard:
+                break
+            elif move.message == Message.NoClaim and move.notifying:
+                noClaimCount += 1
+                if noClaimCount == 2:
+                    # self.game.debug('everybody said "I am not interested", so %s claims chow now' %
+                    #     self.game.myself.name)
+                    return result
+            elif move.message in (Message.Pung, Message.Kong) and move.notifying:
+                # self.game.debug('somebody said Pung or Kong, so %s suppresses Chow' % self.game.myself.name)
+                return Message.NoClaim
+        if delay < self.game.ruleset.claimTimeout * 0.95:
+            # one of those slow humans is still thinking
+            return deferLater(reactor, delayStep, self.__delayAnswer, result, delay, delayStep)
+        # self.game.debug('%s must chow now because timeout is over' % self.game.myself.name)
+        return result
+
     def ask(self, move, answers):
         """this is where the robot AI should go.
         sends answer and one parameter to server"""
         delay = 0.0
         delayStep = 0.1
-        def delayed(result, delay):
-            """try again, may we chow now?"""
-            noClaimCount = 0
-            delay += delayStep
-            for move in self.game.lastMoves():
-                # latest move first
-                if move.message == Message.Discard:
-                    break
-                elif move.message == Message.PopupMsg and move.msg == 'No Claim':
-                    noClaimCount += 1
-                    if noClaimCount == 2:
-                        # everybody said "I am not interested", so we claim chow now
-                        return result
-                elif move.message == Message.PopupMsg and move.msg in ('Pung', 'Kong'):
-                    # somebody said Pung or Kong, so we suppress our Chow
-                    return
-            if delay < self.game.ruleset.claimTimeout * 0.95:
-                # one of those slow humans is still thinking
-                return deferLater(reactor, delayStep, delayed, result, delay)
-            return result
-        self.computeSayable(move, answers)
+        self._computeSayable(move, answers)
         result = self.intelligence.selectAnswer(answers)
         if result[0] == Message.Chow:
-            # wait to see if somebody says Pung or Kong
-            return deferLater(reactor, delayStep, delayed, result, delay)
+            # self.game.debug('%s waits to see if somebody says Pung or Kong before saying chow' %
+            #     self.game.myself.name)
+            return deferLater(reactor, delayStep, self.__delayAnswer, result, delay, delayStep)
         return succeed(result)
 
     def thatWasMe(self, player):
@@ -324,14 +336,23 @@ class Client(pb.Referenceable):
             return False
         return player == self.game.myself
 
+    @staticmethod
+    def __jellyMessage(value):
+        """the Message classes are not pb.copyable, convert them into their names"""
+        if value is None:
+            return Message.OK.name
+        else:
+            return Message.jelly(value, value)
+
     def remote_move(self, playerName, command, *dummyArgs, **kwargs):
         """the server sends us info or a question and always wants us to answer"""
-        player = None
         if self.game:
-            if not self.game.client:
-                # we aborted the game, ignore what the server tells us
-                return
             player = self.game.playerByName(playerName)
+        elif playerName:
+            player = PlayingPlayer(None)
+            player.name = playerName
+        else:
+            player = None
         move = Move(player, command, kwargs)
         if Debug.traffic:
             if self.isHumanClient():
@@ -345,41 +366,45 @@ class Client(pb.Referenceable):
                 if move.token != self.game.handId(withAI=False):
                     logException( 'wrong token: %s, we have %s' % (move.token, self.game.handId()))
         with Duration('Move %s:' % move):
-            return self.exec_move(move)
-
-    @staticmethod
-    def convertMessage(answer, answer2=None):
-        """the client is done with executing the move. Animations have ended.
-        Now we convert Message objects to their name for the write transfer.
-        This callback may be called either on the Deferred representing
-        the answer. In that case, parameter "answer" is used.
-        Or it may be called as a callback on something else like animate().
-        In that case, we use answer2."""
-        if answer2 is not None:
-            answer = answer2
-        if not isinstance(answer, Deferred):
-            if isinstance(answer, Message):
-                answer = answer.name
-            if isinstance(answer, tuple) and isinstance(answer[0], Message):
-                answer = tuple(list([answer[0].name] + list(answer[1:])))
-        return answer
+            return self.exec_move(move).addCallback(self.__jellyMessage)
 
     def exec_move(self, move):
         """mirror the move of a player as told by the the game server"""
-        answer = move.message.clientAction(self, move)
+        message = move.message
+        if message.needsGame and not self.game:
+            # server already disconnected, see HumanClient.remote_ServerDisconnects
+            return succeed(Message.OK)
+        action = message.notifyAction if move.notifying else message.clientAction
+        answer = action(self, move)
         if not isinstance(answer, Deferred):
             answer = succeed(answer)
-        answer.addCallback(self.convertMessage)
-        if self.game:
-            if move.player and not move.player.scoreMatchesServer(move.score):
-                self.game.close()
-            self.game.moves.append(move)
-        if move.message == Message.Discard:
+        game = self.game
+        if game:
+            if not move.notifying and move.player and not move.player.scoreMatchesServer(move.score):
+                game.close()
+            game.moves.append(move)
+# This is an example how to find games where specific situations arise. We prefer games where this
+# happens very early for easier reproduction. So set number of rounds to 1 in the ruleset before doing this.
+# This example looks for a situation where the single human player may call Chow but one of the
+# robot players calls Pung. See https://bugs.kde.org/show_bug.cgi?id=318981
+#            if self.isHumanClient() and game.nextPlayer() == game.myself:
+#                # I am next
+#                if message == Message.Pung and move.notifying:
+#                    # somebody claimed a pung
+#                    if move.player != game.myself:
+#                        # it was not me
+#                        if game.handctr == 0 and len(game.moves) < 30:
+#                            # early on in the game
+#                            if self.__maySayChow():
+#                                # I may say Chow
+#                                print('FOUND EXAMPLE IN:', game.handId(withMoveCount=True))
+
+        if message == Message.Discard:
             # do not block here, we want to get the clientDialog
             # before the animated tile reaches its end position
             animate()
             return answer
-        elif move.message == Message.AskForClaims:
+        elif message == Message.AskForClaims:
             # no need to start an animation. If we did the below standard clause, this is what
             # could happen:
             # 1. user says Chow
@@ -390,29 +415,32 @@ class Client(pb.Referenceable):
         else:
             # return answer only after animation ends. Put answer into
             # the Deferred returned by animate().
-            return animate().addCallback(self.convertMessage, answer)
+            return animate().addCallback(lambda x: answer)
 
     def claimed(self, move):
         """somebody claimed a discarded tile"""
-        calledTile = self.game.lastDiscard
+        if Internal.field:
+            calledTileItem = Internal.field.discardBoard.lastDiscarded
+            calledTile = calledTileItem.tile
+            Internal.field.discardBoard.lastDiscarded = None
+        else:
+            calledTileItem = None
+            calledTile = self.game.lastDiscard
         self.game.lastDiscard = None
-        calledTileName = calledTile.element
-        self.game.discardedTiles[calledTileName.lower()] -= 1
-        assert calledTileName in move.source, '%s %s'% (calledTileName, move.source)
-        if InternalParameters.field:
-            InternalParameters.field.discardBoard.lastDiscarded = None
-        move.player.lastTile = calledTileName.lower()
+        self.game.discardedTiles[calledTile.lower()] -= 1
+        assert calledTile in move.meld, '%s %s'% (calledTile, move.meld)
+        move.player.lastTile = calledTile.lower()
         move.player.lastSource = 'd'
-        hadTiles = move.source[:]
-        hadTiles.remove(calledTileName)
+        hadTiles = move.meld[:]
+        hadTiles.remove(calledTile)
         if not self.thatWasMe(move.player) and not self.game.playOpen:
             move.player.showConcealedTiles(hadTiles)
-        move.exposedMeld = move.player.exposeMeld(hadTiles, calledTile=calledTile)
+        move.exposedMeld = move.player.exposeMeld(hadTiles, calledTile=calledTileItem or calledTile)
         if self.thatWasMe(move.player):
             if move.message != Message.Kong:
                 # we will get a replacement tile first
                 return self.myAction(move)
-        elif self.game.prevActivePlayer == self.game.myself and self.perspective:
+        elif self.game.prevActivePlayer == self.game.myself and self.connection:
             # even here we ask: if our discard is claimed we need time
             # to notice - think 3 robots or network timing differences
             return self.ask(move, [Message.OK])
@@ -450,10 +478,10 @@ class Client(pb.Referenceable):
         """returns answer arguments for the server if calling pung is possible.
         returns the meld to be completed"""
         if self.game.lastDiscard:
-            element = self.game.lastDiscard.element
-            assert element[0].isupper(), str(self.game.lastDiscard)
-            if self.game.myself.concealedTileNames.count(element) >= 2:
-                return [element] * 3
+            lastDiscard = self.game.lastDiscard
+            assert lastDiscard[0].isupper(), lastDiscard
+            if self.game.myself.concealedTileNames.count(lastDiscard) >= 2:
+                return [lastDiscard] * 3
 
     def __maySayKong(self):
         """returns answer arguments for the server if calling or declaring kong is possible.
@@ -466,11 +494,11 @@ class Client(pb.Referenceable):
         myself = game.myself
         robbableTile = withDiscard = None
         if move.message == Message.DeclaredKong:
-            withDiscard = move.source[0].capitalize()
+            withDiscard = move.tiles[0].upper()
             if move.player != myself:
-                robbableTile = move.exposedMeld.pairs[1] # we want it capitalized for a hidden Kong
+                robbableTile = move.exposedMeld[1] # we want it capitalized for a hidden Kong
         elif move.message == Message.AskForClaims:
-            withDiscard = game.lastDiscard.element
+            withDiscard = game.lastDiscard
         hand = myself.computeHand(withTile=withDiscard, robbedTile=robbableTile, asWinner=True)
         if hand.won:
             if Debug.robbingKong:
@@ -480,7 +508,7 @@ class Client(pb.Referenceable):
             if Debug.mahJongg:
                 game.debug('%s may say MJ:%s, active=%s' % (
                     myself, list(x for x in game.players), game.activePlayer))
-            return (meldsContent(hand.hiddenMelds), withDiscard, list(hand.lastMeld.pairs))
+            return (meldsContent(hand.hiddenMelds), withDiscard, hand.lastMeld)
 
     def __maySayOriginalCall(self):
         """returns True if Original Call is possible"""
@@ -491,7 +519,7 @@ class Client(pb.Referenceable):
                     self.game.debug('%s may say Original Call' % myself)
                 return True
 
-    def computeSayable(self, move, answers):
+    def _computeSayable(self, move, answers):
         """find out what the player can legally say with this hand"""
         self.sayable = {}
         for message in Message.defined.values():
