@@ -21,9 +21,12 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 Read the user manual for a description of the interface to this scoring engine
 """
 
+from tile import elements
 from meld import Meld, CONCEALED, EXPOSED, CLAIMEDKONG, REST, elementKey
-from common import elements, IntDict
+from common import IntDict, WINDS
 from message import Message
+from query import Query
+from util import logDebug
 
 class Function(object):
     """Parent for all Function classes. We need to implement
@@ -35,7 +38,13 @@ class Function(object):
     def __init__(self):
         self.options = {}
 
-# pylint: disable=C0111
+    def __str__(self):
+        return self.__class__.__name__
+
+    def __repr__(self):
+        return self.__class__.__name__
+
+# pylint: disable=missing-docstring
 # the class and method names are mostly self explaining, we do not
 # need docstringss
 
@@ -43,116 +52,116 @@ class DragonPungKong(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
         return (len(meld) >= 3
-            and meld.pairs[0][0].lower() == 'd'
+            and meld[0][0].lower() == 'd'
             and (meld.isPung() or meld.isKong()))
 
 class RoundWindPungKong(Function):
     @staticmethod
     def appliesToMeld(hand, meld):
-        return len(meld) >= 3 and meld.pairs[0].lower() == 'w' + hand.roundWind
+        return len(meld) >= 3 and meld[0].lower() == 'w' + hand.roundWind
 
 class ExposedMinorPung(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return meld.isPung() and meld.pairs.isLower(0, 3) and meld.pairs[0][1] in '2345678'
+        return meld.isPung() and meld.isLower(0, 3) and meld[0][1] in '2345678'
 
 class ExposedTerminalsPung(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return meld.isPung() and meld.pairs.isLower(0, 3) and meld.pairs[0][1] in '19'
+        return meld.isPung() and meld.isLower(0, 3) and meld[0][1] in '19'
 
 class ExposedHonorsPung(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return meld.isPung() and meld.pairs[0][0] in 'wd'
+        return meld.isPung() and meld[0][0] in 'wd'
 
 class ExposedMinorKong(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 4 and meld.pairs.isLower(0, 3) and meld.pairs[0][1] in '2345678'
+        return len(meld) == 4 and meld.isLower(0, 3) and meld[0][1] in '2345678'
 
 class ExposedTerminalsKong(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 4 and meld.pairs.isLower(0, 3) and meld.pairs[0][1] in '19'
+        return len(meld) == 4 and meld.isLower(0, 3) and meld[0][1] in '19'
 
 class ExposedHonorsKong(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 4 and meld.pairs.isLower(0, 3) and meld.pairs[0][0] in 'wd'
+        return len(meld) == 4 and meld.isLower(0, 3) and meld[0][0] in 'wd'
 
 class ConcealedMinorPung(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return meld.isPung() and meld.pairs.isUpper(0, 3) and meld.pairs[0][1] in '2345678'
+        return meld.isPung() and meld.isUpper(0, 3) and meld[0][1] in '2345678'
 
 class ConcealedTerminalsPung(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return meld.isPung() and meld.pairs.isUpper(0, 3) and meld.pairs[0][1] in '19'
+        return meld.isPung() and meld.isUpper(0, 3) and meld[0][1] in '19'
 
 class ConcealedHonorsPung(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return meld.isPung() and meld.pairs[0][0] in 'WD'
+        return meld.isPung() and meld[0][0] in 'WD'
 
 class ConcealedMinorKong(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 4 and meld.state == CONCEALED and meld.pairs[0][1] in '2345678'
+        return len(meld) == 4 and meld.state == CONCEALED and meld[0][1] in '2345678'
 
 class ConcealedTerminalsKong(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 4 and meld.state == CONCEALED and meld.pairs[0][1] in '19'
+        return len(meld) == 4 and meld.state == CONCEALED and meld[0][1] in '19'
 
 class ConcealedHonorsKong(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 4 and meld.state == CONCEALED and meld.pairs[0][0] in 'wd'
+        return len(meld) == 4 and meld.state == CONCEALED and meld[0][0] in 'wd'
 
 class OwnWindPungKong(Function):
     @staticmethod
     def appliesToMeld(hand, meld):
-        return len(meld) >= 3 and meld.pairs[0].lower() == 'w' + hand.ownWind
+        return len(meld) >= 3 and meld[0].lower() == 'w' + hand.ownWind
 
 class OwnWindPair(Function):
     @staticmethod
     def appliesToMeld(hand, meld):
-        return len(meld) == 2 and meld.pairs[0].lower() == 'w' + hand.ownWind
+        return len(meld) == 2 and meld[0].lower() == 'w' + hand.ownWind
 
 class RoundWindPair(Function):
     @staticmethod
     def appliesToMeld(hand, meld):
-        return len(meld) == 2 and meld.pairs[0].lower() == 'w' + hand.roundWind
+        return len(meld) == 2 and meld[0].lower() == 'w' + hand.roundWind
 
 class DragonPair(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 2 and meld.pairs[0][0].lower() == 'd'
+        return len(meld) == 2 and meld[0][0].lower() == 'd'
 
 class LastTileCompletesPairMinor(Function):
     @staticmethod
     def appliesToHand(hand):
         return (hand.lastMeld and len(hand.lastMeld) == 2
-            and hand.lastMeld.pairs[0][0] == hand.lastMeld.pairs[1][0]
+            and hand.lastMeld[0][0] == hand.lastMeld[1][0]
             and hand.lastTile[1] in '2345678')
 
 class Flower(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 1 and meld.pairs[0][0] == 'f'
+        return len(meld) == 1 and meld[0][0] == 'f'
 
 class Season(Function):
     @staticmethod
     def appliesToMeld(dummyHand, meld):
-        return len(meld) == 1 and meld.pairs[0][0] == 'y'
+        return len(meld) == 1 and meld[0][0] == 'y'
 
 class LastTileCompletesPairMajor(Function):
     @staticmethod
     def appliesToHand(hand):
         return (hand.lastMeld and len(hand.lastMeld) == 2
-            and hand.lastMeld.pairs[0][0] == hand.lastMeld.pairs[1][0]
+            and hand.lastMeld[0][0] == hand.lastMeld[1][0]
             and hand.lastTile[1] not in '2345678')
 
 class LastFromWall(Function):
@@ -377,7 +386,7 @@ class TripleKnitting(Function):
     def shouldTry(self, hand, maxMissing=3):
         if hand.declaredMelds:
             return False
-        tripleWanted = 7 - maxMissing / 3 # count triples
+        tripleWanted = 7 - maxMissing // 3 # count triples
         tripleCount = len(self.findTriples(hand)[0])
         return tripleCount >= tripleWanted
 
@@ -427,7 +436,7 @@ class Knitting(Function):
     def shouldTry(self, hand, maxMissing=4):
         if hand.declaredMelds:
             return False
-        pairWanted = 7 - maxMissing / 2 # count pairs
+        pairWanted = 7 - maxMissing // 2 # count pairs
         pairCount = len(self.findCouples(hand)[0])
         return pairCount >= pairWanted
 
@@ -542,10 +551,8 @@ class AllPairHonors(Function):
                 pairCount += 1
             elif count == 4:
                 kongCount += 1
-        pairWanted = 7 - maxMissing / 2 # count pairs
+        pairWanted = 7 - maxMissing // 2 # count pairs
         result = pairCount >= pairWanted or (pairCount + kongCount * 2) > pairWanted
-        if pairCount > 5:
-            hand.debug('have %d pairs for allpairhonors: %s' % (pairCount, hand.tileNames))
         return result
     @staticmethod
     def rearrange(dummyHand, pairs):
@@ -608,7 +615,7 @@ class LittleThreeDragons(Function):
 class FourBlessingsHoveringOverTheDoor(Function):
     @staticmethod
     def appliesToHand(hand):
-        return len([x for x in hand.melds if len(x) >= 3 and x.pairs[0][0] in 'wW']) == 4
+        return len([x for x in hand.melds if len(x) >= 3 and x[0][0] in 'wW']) == 4
 
 class AllGreen(Function):
     @staticmethod
@@ -678,6 +685,39 @@ class ScratchingPole(Function):
     def appliesToHand(hand):
         return RobbingKong.appliesToHand(hand) and hand.lastTile == 'b2'
 
+class StandardRotation(Function):
+    @staticmethod
+    def rotate(game):
+        return game.winner and game.winner.wind != 'E'
+
+class EastWonNineTimesInARow(Function):
+    nineTimes = 9
+    @staticmethod
+    def appliesToHand(hand):
+        if not hand.player:
+            return False
+        game = hand.player.game
+        return EastWonNineTimesInARow.appliesToGame(game)
+    @staticmethod
+    def appliesToGame(game, needWins=None):
+        if needWins is None:
+            needWins = EastWonNineTimesInARow.nineTimes
+            if game.isScoringGame():
+                # we are only proposing for the last needed Win
+                needWins  -= 1
+        if game.winner and game.winner.wind == 'E' and game.notRotated >= needWins:
+            prevailing = WINDS[game.roundsFinished % 4]
+            eastMJCount = int(Query("select count(1) from score "
+                "where game=%d and won=1 and wind='E' and player=%d "
+                "and prevailing='%s'" % \
+                (game.gameid, game.players['E'].nameid, prevailing)).records[0][0])
+            return eastMJCount == needWins
+        return False
+    @staticmethod
+    def rotate(game):
+        return EastWonNineTimesInARow.appliesToGame(game, needWins = EastWonNineTimesInARow.nineTimes)
+
+
 class StandardMahJongg(Function):
     @staticmethod
     def computeLastMelds(hand):
@@ -689,12 +729,12 @@ class StandardMahJongg(Function):
             checkMelds = hand.hiddenMelds
         else:
             checkMelds = hand.declaredMelds
-        return [x for x in checkMelds if hand.lastTile in x.pairs and len(x) < 4]
+        return [x for x in checkMelds if hand.lastTile in x and len(x) < 4]
 
     @staticmethod
     def appliesToHand(hand):
         """winner rules are not yet applied to hand"""
-        # pylint: disable=R0911
+        # pylint: disable=too-many-return-statements
         # too many return statements
         if len(hand.melds) != 5:
             return False
@@ -714,31 +754,33 @@ class StandardMahJongg(Function):
         return hand.score.doubles + doublingWinnerRules >= hand.ruleset.minMJDoubles
 
     @staticmethod
-    def fillChow(color, values):
+    def fillChow(group, values):
         val0, val1 = values
         if val0 + 1 == val1:
             if val0 == 1:
-                return set([color + str(val1 + 1)])
+                return set([group + str(val1 + 1)])
             if val0 == 8:
-                return set([color + str(val0 - 1)])
-            return set([color + str(val0 - 1), color + str(val0 + 2)])
+                return set([group + str(val0 - 1)])
+            return set([group + str(val0 - 1), group + str(val0 + 2)])
         else:
-            assert val0 + 2 == val1, 'color:%s values:%s' % (color, values)
-            return set([color + str(val0 + 1)])
+            assert val0 + 2 == val1, 'group:%s values:%s' % (group, values)
+            return set([group + str(val0 + 1)])
 
     @staticmethod
     def winningTileCandidates(hand):
-        # pylint: disable=R0914,R0911,R0912,R0915
+        # pylint: disable=too-many-locals,too-many-return-statements,too-many-branches,too-many-statements
         if len(hand.melds) > 7:
             # hope 7 is sufficient, 6 was not
             return set()
-        inHand = list(x.lower() for x in hand.inHand)
-        if not hand.inHand:
+        if not hand.tilesInHand:
             return set()
+        inHand = list(x.lower() for x in hand.tilesInHand)
         result = inHand[:]
         pairs = 0
         isolated = 0
         maxChows = hand.ruleset.maxChows - sum(x.isChow() for x in hand.declaredMelds)
+        # TODO: does not differentiate between maxChows == 1 and maxChows > 1
+        # test with kajonggtest and a ruleset where maxChows == 2
         if maxChows < 0:
             return set()
         if maxChows == 0:
@@ -764,8 +806,8 @@ class StandardMahJongg(Function):
         if maxChows == 0:
             return set(result)
         melds = []
-        for color in hand.suits & set('sbc'):
-            values = sorted(int(x[1]) for x in result if x[0] == color)
+        for group in hand.suits & set('sbc'):
+            values = sorted(int(x[1]) for x in result if x[0] == group)
             changed = True
             while (changed and len(values) > 2
                     and values.count(values[0]) == 1
@@ -774,15 +816,15 @@ class StandardMahJongg(Function):
                 changed = False
                 if values[0] + 2 == values[2] and (len(values) == 3 or values[3] > values[0] + 3):
                     # print('removing first 3 from %s' % values)
-                    meld = Meld([color + str(values[x]) for x in range(3)])
-                    for pair in meld.pairs:
+                    meld = Meld([group + str(values[x]) for x in range(3)])
+                    for pair in meld:
                         result.remove(pair)
                     melds.append(meld)
                     values = values[3:]
                     changed = True
                 elif values[0] + 1 == values[1] and values[2] > values[0] + 2:
                     # print('found incomplete chow at start of %s' % values)
-                    return StandardMahJongg.fillChow(color, values[:2])
+                    return StandardMahJongg.fillChow(group, values[:2])
             changed = True
             while (changed and len(values) > 2
                     and values.count(values[-1]) == 1
@@ -790,34 +832,33 @@ class StandardMahJongg(Function):
                     and values.count(values[-3]) == 1):
                 changed = False
                 if values[-1] - 2 == values[-3] and (len(values) == 3 or values[-4] < values[-1] - 3):
-                    # print('removing last 3 from %s' % values)
-                    meld = Meld([color + str(values[x]) for x in range(-3, 0)])
-                    for pair in meld.pairs:
+                    meld = Meld([group + str(values[x]) for x in range(-3, 0)])
+                    for pair in meld:
                         result.remove(pair)
                     melds.append(meld)
                     values = values[:-3]
                     changed = True
                 elif values[-1] - 1 == values[-2] and values[-3] < values[-1] - 2:
                     # print('found incomplete chow at end of %s' % values)
-                    return StandardMahJongg.fillChow(color, values[-2:])
+                    return StandardMahJongg.fillChow(group, values[-2:])
 
             if len(values) % 3 == 0:
                 # adding a 4th, 7th or 10th tile with this color can not let us win,
                 # so we can exclude this color from the candidates
-                result = list(x for x in result if x[0] != color)
+                result = list(x for x in result if x[0] != group)
                 continue
             valueSet = set(values)
             if len(values) == 4 and len(values) == len(valueSet):
                 if values[0] + 3 == values[-1]:
-                    # print('seq4 in %s' % hand.inHand)
-                    return set([color + str(values[0]), color + str(values[-1])])
+                    # print('seq4 in %s' % hand.tilesInHand)
+                    return set([group + str(values[0]), group + str(values[-1])])
             if len(values) == 7 and len(values) == len(valueSet):
                 if values[0] + 6 == values[6]:
-                    # print('seq7 in %s' % hand.inHand)
-                    return set([color + str(values[0]), color + str(values[3]), color + str(values[6])])
+                    # print('seq7 in %s' % hand.tilesInHand)
+                    return set([group + str(values[0]), group + str(values[3]), group + str(values[6])])
             if len(values) == 1:
                 # only a pair of this value is possible
-                return set([color.upper() + str(values[0])])
+                return set([group.upper() + str(values[0])])
             if len(valueSet) == 1:
                 # no chow reachable, only pair/pung
                 continue
@@ -833,17 +874,17 @@ class StandardMahJongg(Function):
                 if maxChows == 0:
                     # not a calling hand
                     return set()
-                return StandardMahJongg.fillChow(color, values)
+                return StandardMahJongg.fillChow(group, values)
             if (len(values) == 4 and len(valueSet) == 2
                     and values[0] == values[1] and values[2] == values[3]):
-                # print('we have 2 pairs of %s' % color)
-                return set([color + str(values[0]), color + str(values[2])])
+                # print('we have 2 pairs of %s' % group)
+                return set([group + str(values[0]), group + str(values[2])])
             if maxChows:
                 for value in valueSet:
                     if value > 1:
-                        result.append(color + str(value - 1))
+                        result.append(group + str(value - 1))
                     if value < 9:
-                        result.append(color + str(value + 1))
+                        result.append(group + str(value + 1))
         return set(result)
 
     @staticmethod
@@ -860,12 +901,12 @@ class StandardMahJongg(Function):
         _ = [pair for pair in pairs if pair[0] in 'DW']
         honourResult = hand.splitRegex(''.join(_)) # easy since they cannot have a chow
         splitVariants = {}
-        for color in 'SBC':
-            colorPairs = [pair for pair in pairs if pair[0] == color]
-            if not colorPairs:
-                splitVariants[color] = [None]
+        for group in 'SBC':
+            groupPairs = [pair for pair in pairs if pair[0] == group]
+            if not groupPairs:
+                splitVariants[group] = [None]
                 continue
-            splitVariants[color] = hand.genVariants(colorPairs)
+            splitVariants[group] = hand.genVariants(groupPairs)
         bestHand = None
         bestVariant = None
         for combination in ((s, b, c)
@@ -942,7 +983,7 @@ class ThirteenOrphans(Function):
 
     @staticmethod
     def computeLastMelds(hand):
-        meldSize = hand.inHand.count(hand.lastTile)
+        meldSize = hand.tilesInHand.count(hand.lastTile)
         return [Meld([hand.lastTile] * meldSize)]
 
     @staticmethod
@@ -1037,13 +1078,13 @@ class ThirteenOrphans(Function):
 class OwnFlower(Function):
     @staticmethod
     def appliesToHand(hand):
-        fsPairs = list(x.pairs[0] for x in hand.bonusMelds)
+        fsPairs = list(x[0] for x in hand.bonusMelds)
         return 'f' + hand.ownWind in fsPairs
 
 class OwnSeason(Function):
     @staticmethod
     def appliesToHand(hand):
-        fsPairs = list(x.pairs[0] for x in hand.bonusMelds)
+        fsPairs = list(x[0] for x in hand.bonusMelds)
         return 'y' + hand.ownWind in fsPairs
 
 class OwnFlowerOwnSeason(Function):
@@ -1055,12 +1096,12 @@ class OwnFlowerOwnSeason(Function):
 class AllFlowers(Function):
     @staticmethod
     def appliesToHand(hand):
-        return len([x for x in hand.bonusMelds if x.pairs[0][0] == 'f']) == 4
+        return len([x for x in hand.bonusMelds if x[0][0] == 'f']) == 4
 
 class AllSeasons(Function):
     @staticmethod
     def appliesToHand(hand):
-        return len([x for x in hand.bonusMelds if x.pairs[0][0] == 'y']) == 4
+        return len([x for x in hand.bonusMelds if x[0][0] == 'y']) == 4
 
 class ThreeConcealedPongs(Function):
     @staticmethod
