@@ -18,12 +18,14 @@ along with this program if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
-from common import InternalParameters, Preferences, ZValues
+from common import Internal, Preferences, ZValues
 from PyQt4.QtCore import QRectF, QPointF
 from PyQt4.QtGui import QGraphicsSimpleTextItem
 
 from board import PlayerWind, YellowText, Board, rotateCenter
-from game import Wall
+from wall import Wall, KongBox
+from tile import Tile
+from uitile import UITile
 from animation import animate, afterCurrentAnimationDo, Animated, \
     ParallelAnimationGroup
 
@@ -33,15 +35,14 @@ class UIWallSide(Board):
         Board.__init__(self, length, 1, tileset, boardRotation=boardRotation)
         self.length = length
 
-    # pylint: disable=R0201
     def name(self):
         """name for debug messages"""
-        game = InternalParameters.field.game
+        game = Internal.field.game
         if not game:
             return 'NOGAME'
         for player in game.players:
             if player.front == self:
-                return 'wallside %s'% player.name
+                return 'wall %s'% player.name
 
     def center(self):
         """returns the center point of the wall in relation to the faces of the upper level"""
@@ -51,23 +52,53 @@ class UIWallSide(Board):
         result.setX(result.x() + faceRect.height()/2) # corner tile
         return result
 
+    def hide(self):
+        """hide all my parts"""
+        self.windTile.hide()
+        self.nameLabel.hide()
+        Board.hide(self)
+
+class UIKongBox(KongBox):
+    """Kong box with UITiles"""
+    # pylint: disable=incomplete-protocol
+    def __init__(self):
+        KongBox.__init__(self)
+
+    def fill(self, tiles):
+        """fill the box"""
+        for tile in self._tiles:
+            tile.cross = False
+        KongBox.fill(self, tiles)
+        for tile in self._tiles:
+            tile.cross = True
+
+    def pop(self, count):
+        """get count tiles from kong box"""
+        result = KongBox.pop(self, count)
+        for tile in result:
+            tile.cross = False
+        return result
+
 class UIWall(Wall):
     """represents the wall with four sides. self.wall[] indexes them counter clockwise, 0..3. 0 is bottom."""
+    tileClass = UITile
+    kongBoxClass = UIKongBox
+
     def __init__(self, game):
         """init and position the wall"""
         # we use only white dragons for building the wall. We could actually
         # use any tile because the face is never shown anyway.
         game.wall = self
         Wall.__init__(self, game)
-        self.__square = Board(1, 1, InternalParameters.field.tileset)
+        self.__square = Board(1, 1, Internal.field.tileset)
         self.__square.setZValue(ZValues.marker)
         sideLength = len(self.tiles) // 8
-        self.__sides = [UIWallSide(InternalParameters.field.tileset, boardRotation, sideLength) \
+        self.__sides = [UIWallSide(Internal.field.tileset, boardRotation, sideLength) \
             for boardRotation in (0, 270, 180, 90)]
         for side in self.__sides:
             side.setParentItem(self.__square)
             side.lightSource = self.lightSource
-            side.windTile = PlayerWind('E', InternalParameters.field.windTileset, parent=side)
+            side.windTile = PlayerWind('E', Internal.field.windTileset, parent=side)
             side.windTile.hide()
             side.nameLabel = QGraphicsSimpleTextItem('', side)
             font = side.nameLabel.font()
@@ -82,10 +113,10 @@ class UIWall(Wall):
         self.__sides[2].setPos(xHeight=1, xWidth=sideLength, yHeight=1)
         self.__sides[1].setPos(xWidth=sideLength, yWidth=sideLength, yHeight=1 )
         self.showShadows = Preferences.showShadows
-        InternalParameters.field.centralScene.addItem(self.__square)
+        Internal.field.centralScene.addItem(self.__square)
 
-    # pylint: disable=R0201
-    def name(self):
+    @staticmethod
+    def name():
         """name for debug messages"""
         return 'wall'
 
@@ -107,24 +138,20 @@ class UIWall(Wall):
 
     def hide(self):
         """hide all four walls and their decorators"""
+        self.living = []
+        self.kongBox.fill([])
         for side in self.__sides:
-            side.windTile.hide()
-            side.nameLabel.hide()
             side.hide()
-            del side
-        for tile in self.tiles:
-            if tile.graphics:
-                tile.graphics.hide()
-            del tile
         self.tiles = []
-        InternalParameters.field.centralScene.removeItem(self.__square)
+        Internal.field.centralScene.removeItem(self.__square)
 
     def __shuffleTiles(self):
         """shuffle tiles for next hand"""
-        discardBoard = InternalParameters.field.discardBoard
+        discardBoard = Internal.field.discardBoard
         places = [(x, y) for x in range(-3, discardBoard.width+3) for y in range(-3, discardBoard.height+3)]
         places = self.game.randomGenerator.sample(places, len(self.tiles))
         for idx, tile in enumerate(self.tiles):
+            assert isinstance(tile, UITile)
             tile.dark = True
             tile.setBoard(discardBoard, *places[idx])
 
@@ -132,9 +159,9 @@ class UIWall(Wall):
         """builds the wall without dividing"""
         # recycle used tiles
         for tile in self.tiles:
-            tile.element = 'Xy'
+            tile.tile = Tile('Xy')
             tile.dark = True
-#        field = InternalParameters.field
+#        field = Internal.field
 #        animateBuild = not field.game.isScoringGame() and not self.game.isFirstHand()
         animateBuild = False
         with Animated(animateBuild):
@@ -156,61 +183,56 @@ class UIWall(Wall):
         self.__setDrawingOrder()
         return animate()
 
-    @apply
-    def lightSource():
-        """setting this actually changes the visuals. For
-        possible values see LIGHTSOURCES"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self.__square.lightSource
-        def fset(self, lightSource):
-            # pylint: disable=W0212
-            if self.lightSource != lightSource:
-                assert ParallelAnimationGroup.current is None
-                self.__square.lightSource = lightSource
-                for side in self.__sides:
-                    side.lightSource = lightSource
-                self.__setDrawingOrder()
-        return property(**locals())
+    @property
+    def lightSource(self):
+        """For possible values see LIGHTSOURCES"""
+        return self.__square.lightSource
 
-    @apply
-    def tileset():
-        """setting this actually changes the visuals."""
-        def fget(self):
-            # pylint: disable=W0212
-            return self.__square.tileset
-        def fset(self, value):
-            # pylint: disable=W0212
-            if self.tileset != value:
-                assert ParallelAnimationGroup.current is None
-                self.__square.tileset = value
-                for side in self.__sides:
-                    side.tileset = value
-                self.__resizeHandBoards()
-        return property(**locals())
+    @lightSource.setter
+    def lightSource(self, lightSource):
+        """setting this actually changes the visuals"""
+        if self.lightSource != lightSource:
+            assert ParallelAnimationGroup.current is None
+            self.__square.lightSource = lightSource
+            for side in self.__sides:
+                side.lightSource = lightSource
+            self.__setDrawingOrder()
 
-    @apply
-    # pylint: disable=E0202
-    def showShadows():
+    @property
+    def tileset(self):
+        """The tileset of this wall"""
+        return self.__square.tileset
+
+    @tileset.setter
+    def tileset(self, value):
         """setting this actually changes the visuals."""
-        def fget(self):
-            # pylint: disable=W0212
-            return self.__square.showShadows
-        def fset(self, showShadows):
-            # pylint: disable=W0212
-            if self.showShadows != showShadows:
-                assert ParallelAnimationGroup.current is None
-                self.__square.showShadows = showShadows
-                for side in self.__sides:
-                    side.showShadows = showShadows
-                self.__resizeHandBoards()
-        return property(**locals())
+        if self.tileset != value:
+            assert ParallelAnimationGroup.current is None
+            self.__square.tileset = value
+            for side in self.__sides:
+                side.tileset = value
+            self.__resizeHandBoards()
+
+    @property
+    def showShadows(self):
+        """The current value"""
+        return self.__square.showShadows
+
+    @showShadows.setter
+    def showShadows(self, showShadows):
+        """setting this actually changes the visuals."""
+        if self.showShadows != showShadows:
+            assert ParallelAnimationGroup.current is None
+            self.__square.showShadows = showShadows
+            for side in self.__sides:
+                side.showShadows = showShadows
+            self.__resizeHandBoards()
 
     def __resizeHandBoards(self, dummyResults=None):
         """we are really calling _setRect() too often. But at least it works"""
         for player in self.game.players:
             player.handBoard.computeRect()
-        InternalParameters.field.adjustView()
+        Internal.field.adjustView()
 
     def __setDrawingOrder(self, dummyResults=None):
         """set drawing order of the wall"""
@@ -218,7 +240,7 @@ class UIWall(Wall):
         for idx, side in enumerate(self.__sides):
             side.level = (levels[side.lightSource][idx] + 1) * ZValues.boardLevelFactor
 
-    def _moveDividedTile(self, tile, offset):
+    def __moveDividedTile(self, tile, offset):
         """moves a tile from the divide hole to its new place"""
         board = tile.board
         newOffset = tile.xoffset + offset
@@ -227,8 +249,9 @@ class UIWall(Wall):
             sideIdx = self.__sides.index(tile.board)
             board = self.__sides[(sideIdx+1) % 4]
         tile.setBoard(board, newOffset % sideLength, 0, level=2)
+        tile.update()
 
-    def placeLooseTiles(self):
+    def _placeLooseTiles(self):
         """place the last 2 tiles on top of kong box"""
         assert len(self.kongBox) % 2 == 0
         afterCurrentAnimationDo(self.__placeLooseTiles2)
@@ -239,8 +262,8 @@ class UIWall(Wall):
         if placeCount >= 4:
             first = min(placeCount-1, 5)
             second = max(first-2, 1)
-            self._moveDividedTile(self.kongBox[-1], second)
-            self._moveDividedTile(self.kongBox[-2], first)
+            self.__moveDividedTile(self.kongBox[-1], second)
+            self.__moveDividedTile(self.kongBox[-2], first)
 
     def divide(self):
         """divides a wall, building a living and and a dead end"""
@@ -251,7 +274,7 @@ class UIWall(Wall):
                 # in kongbox in a previous game
                 # might not be there anymore. This gets rid
                 # of the cross on them.
-                tile.graphics.update()
+                tile.update()
             # move last two tiles onto the dead end:
             return animate().addCallback(self.__placeLooseTiles2)
 
@@ -265,11 +288,7 @@ class UIWall(Wall):
         side = player.front
         sideCenter = side.center()
         name = side.nameLabel
-        if player.handBoard:
-            player.newHandContent = player.computeNewHand()
-            name.setText(' - '.join([player.localName, unicode(player.newHandContent.total())]))
-        else:
-            name.setText(player.localName)
+        name.setText(' - '.join([player.localName, unicode(player.handTotalForWall())]))
         name.resetTransform()
         if side.rotation() == 180:
             rotateCenter(name, 180)
